# Copyright (C) 2020 Krishnaswamy Lab, Yale University

import numpy as np
import graphtools as gt
import meld
from utils import make_batches, assert_warns_message, assert_raises_message

def test_utils():
    data, labels = make_batches(n_pts_per_cluster=250)
    G = gt.Graph(data, sample_idx=labels, use_pygsp=True)
    EES = meld.MELD().fit_transform(G, labels)

    clusters = meld.VertexFrequencyCluster().fit_predict(G=G, RES=labels, EES=EES)

    cmap = meld.utils.get_meld_cmap()

    # Test sample normalization
    D = np.random.normal(0, 2, (100, 2))
    # EES wrong shape
    EES = np.ones([100, 3])
    with assert_raises_message(
        ValueError,
            "It is currently only possible to normalize two samaples"
            "from a matched experimental and control pair. As such,"
            "EES.shape[1] must equal 2. Got {}".format(EES.shape[1])
        ):
        meld.utils.normalize_EES_within_replicate(EES)

    # EES correct shape
    EES = np.ones([100, 2])
    meld.utils.normalize_EES_within_replicate(EES=EES)



# Copyright (C) 2020 Krishnaswamy Lab, Yale University

import numpy as np
import pandas as pd
import pygsp
import graphtools
import scprep.utils

from . import utils
from graphtools.estimator import GraphEstimator, attribute
from functools import partial


class MELD(GraphEstimator):
    """MELD operator for filtering signals over a graph.

    Parameters
    ----------
    beta : int, optional, Default: 60
        Amount of smoothing to apply. Default value of 60 determined through analysis
        of simulated data using Splatter (https://github.com/Oshlack/splatter).
    offset: float, optional, Default: 0
        Amount to shift the MELD filter in the eigenvalue spectrum.
        Recommend using an eigenvalue from the graph based on the
        spectral distribution. Should be in interval [0,1]
    order: int, optional, Default: 1
        Falloff and smoothness of the filter.
        High order leads to square-like filters.
    Solver : string, optional, Default: 'chebyshev'
        Method to solve convex problem.
        'chebyshev' uses a chebyshev polynomial approximation of the corresponding filter
        'exact' uses the eigenvalue solution to the problem
    chebyshev_order : int, optional, Default: 50
        Order of chebyshev approximation to use.
    lap_type : ('combinatorial', 'normalized'), Default: 'combinatorial'
        The kind of Laplacian to calculate
    normalize : boolean, optional, Default: True
        If True, the RES is column normalized to sum to 1
    """

    # parameters
    beta = attribute("beta", default=40, on_set=graphtools.utils.check_positive)
    offset = attribute("offset", default=0)
    order = attribute("order", default=1)
    filter = attribute(
        "filter",
        default="heat",
        on_set=partial(graphtools.utils.check_in, ["heat", "laplacian"]),
    )
    solver = attribute(
        "solver",
        default="chebyshev",
        on_set=partial(graphtools.utils.check_in, ["chebyshev", "exact"]),
    )
    chebyshev_order = attribute(
        "chebyshev_order",
        default=30,
        on_set=[graphtools.utils.check_int, graphtools.utils.check_positive],
    )
    lap_type = attribute(
        "lap_type",
        default="combinatorial",
        on_set=partial(graphtools.utils.check_in, ["combinatorial", "normalized"]),
    )

    # stored attributes
    filt = attribute("filt")
    EES = attribute("EES", doc="Enhanced Experimental Signal (smoothed RES)")

    def __init__(
        self,
        beta=60,
        offset=0,
        order=1,
        filter="heat",
        solver="chebyshev",
        chebyshev_order=50,
        lap_type="combinatorial",
        normalize=True,
        anisotropy=1,
        n_landmark=None,
        **kwargs
    ):

        self.beta = beta
        self.offset = offset
        self.order = order
        self.solver = solver
        self.chebyshev_order = chebyshev_order
        self.lap_type = lap_type
        self.filter = filter
        self.normalize = normalize

        kwargs["use_pygsp"] = True
        super().__init__(anisotropy=anisotropy, n_landmark=n_landmark, **kwargs)

    def _reset_graph(self):
        self._reset_filter()

    def _reset_filter(self):
        self.filt = None
        self.EES = None

    def set_params(self, **params):
        for p in [
            "beta",
            "offset",
            "order",
            "solver",
            "chebyshev_order",
            "lap_type",
            "filter",
        ]:
            if p in params and params[p] != getattr(self, p):
                self._reset_filter()
                setattr(self, p, params[p])
                del params[p]
        super().set_params(**params)

    def _RES_from_sample_labels(self):
        '''
        Helper function to take an array of non-numerics and produce an RES.
        '''
        self.sample_labels_ = self.RES
        self.samples = np.unique(self.RES)
        self._RES_cls = pd.DataFrame

        if self.samples.shape[0] == 1:
            # Only have one sample label (i.e. [A, A, A, A])
            self.RES = pd.DataFrame(np.ones(self.sample_labels_.shape[0]),
                                    columns=self.samples)
        elif self.samples.shape[0] == 2:
            # When there's two samples (i.e. [A, A, B, B])
            # LabelBinarizer doesn't work nicely with only two labels
            self.RES = pd.DataFrame([self.sample_labels_ == self.samples[0],
                            self.sample_labels_ == self.samples[1]],
                            dtype=int,
                            index=self.samples).T

        else:
            # We have more than two samples, use label binarizer.
            import sklearn
            self._LB = sklearn.preprocessing.LabelBinarizer()
            RES = self._LB.fit_transform(self.sample_labels_)
            self.RES = pd.DataFrame(RES, columns=self._LB.classes_)
        self._RES_columns = self.RES.columns

        return scprep.utils.toarray(self.RES)

    def transform(self, RES):
        """Filters a signal `RES` over the data graph.

        Parameters
        ----------
        RES : ndarray [n, p]
            1- or 2-dimensional Raw Experimental Signal array to filter.

        Returns
        -------
        EES : ndarray [n, p]
            Enhanced Experimental Signal (smoothed RES)
        """
        self.graph = utils._check_pygsp_graph(self.graph)

        if RES.shape[0] != self.graph.N:
            raise ValueError(
                "Input data ({}) and input graph ({}) "
                "are not of the same size".format(RES.shape, self.graph.N)
            )

        self._RES_cls = type(RES)
        self._RES_index = None
        self._RES_columns = None
        if isinstance(RES, pd.DataFrame):
            self._RES_index, self._RES_columns = RES.index, RES.columns

        self.RES = scprep.utils.toarray(RES)

        # Need to handle multiple cases for how the RES is passed
        # Option 1, a categorical / series / something like ['A', 'A', 'B', 'B']
        if not np.issubdtype(self.RES.dtype, np.number):
            # If we have non-numeric RES, then we should create an RES from it
            self.RES = self._RES_from_sample_labels()

        if self.normalize:
            self.RES = self.RES / self.RES.sum(axis=0)

        self.graph.estimate_lmax()

        # Generate MELD filter
        if self.filter.lower() == "laplacian":

            def filterfunc(x):
                return 1 / (
                    1 + (self.beta * np.abs(x / self.graph.lmax - self.offset)) ** self.order
                )

        elif self.filter.lower() == "heat":

            def filterfunc(x):
                return (
                    np.exp(-self.beta * np.abs(x / self.graph.lmax - self.offset) ** self.order)
                )

        else:
            raise NotImplementedError

        # build filter
        self.filt = pygsp.filters.Filter(self.graph, filterfunc)

        # apply filter
        self.EES = self.filt.filter(self.RES, method=self.solver, order=self.chebyshev_order)

        # normalize if only two samaples
        if len(self.EES.shape) > 1 and self.EES.shape[1] == 2:
            self.EES = utils.normalize_EES_within_replicate(self.EES)

        if self._RES_cls != np.ndarray:
            self.RES = self._RES_cls(self.RES)
            self.EES = self._RES_cls(self.EES)

        if isinstance(self.RES, pd.DataFrame):
            if self._RES_index is not None:
                self.RES.index, self.EES.index = self._RES_index, self._RES_index
            if self._RES_columns is not None:
                self.RES.columns, self.EES.columns = self._RES_columns, self._RES_columns

        return self.EES

    def fit_transform(self, X, RES, **kwargs):
        """Builds the MELD filter over a graph built on data `X` and filters a signal `RES`.

        Parameters
        ----------

        X : array-like, shape=[n_samples, m_features]
            Data on which to build graph to perform data smoothing over.

        RES : array-like, shape=[n_samples, p_signals]
            1- or 2-dimensional Raw Experimental Signal array to filter.

        kwargs : additional arguments for graphtools.Graph

        Returns
        -------
        EES : ndarray, shape=[n_samples, p_signals]
            Enhanced Experimental Signal (smoothed RES)
        """
        self.fit(X, **kwargs)
        self.EES = self.transform(RES)
        return self.EES



# Copyright (C) 2020 Krishnaswamy Lab, Yale University

import numpy as np
import pandas as pd
import graphtools as gt
import meld
import pygsp
import unittest

from scipy import sparse
from parameterized import parameterized
from utils import make_batches, assert_warns_message, assert_raises_message
from nose.tools import assert_raises

from packaging import version


def test_check_pygsp_graph():
    # _check_pygsp_graph
    D = np.random.normal(0, 2, (10, 2))
    G = gt.Graph(D, use_pygsp=False)
    assert isinstance(meld.utils._check_pygsp_graph(G), pygsp.graphs.Graph)
    assert_raises_message(
        TypeError,
        "Input graph should be of type graphtools.base.BaseGraph. "
        "With graphtools, use the `use_pygsp=True` flag.",
        meld.utils._check_pygsp_graph,
        G="hello world",
    )


def test_mnn():
    data, labels = make_batches(n_pts_per_cluster=250)
    meld_op = meld.MELD(verbose=0)
    EES = meld_op.fit_transform(data, labels, sample_idx=labels)
    meld.VertexFrequencyCluster().fit_transform(G=meld_op.graph, RES=labels, EES=EES)


@parameterized([("heat",), ("laplacian",)])
def test_meld(filter):
    # MELD operator
    # Numerical accuracy
    np.random.seed(42)

    def norm(x):
        x = x.copy()
        x = x - np.min(x)
        x = x / np.max(x)
        return x

    D = np.random.normal(0, 2, (1000, 2))
    RES = np.random.binomial(1, norm(D[:, 0]), 1000)

    meld_op = meld.MELD(
        verbose=0, knn=20, decay=10, thresh=0, anisotropy=0,
        filter=filter, solver="exact",
        normalize=False
    )
    B = meld_op.fit_transform(D, RES)

    if version.parse("1.17") <= version.parse(np.__version__) < version.parse("1.18"):
        if meld_op.filter == 'laplacian':
            np.testing.assert_allclose(np.sum(B), 519)
        else:
            np.testing.assert_allclose(np.sum(B), 519)
    else:
        if meld_op.filter == 'laplacian':
            np.testing.assert_allclose(np.sum(B), 532)
        else:
            np.testing.assert_allclose(np.sum(B), 532)

    # check changing filter params resets filter
    meld_op.set_params(beta=meld_op.beta + 1)
    assert meld_op.filt is None
    assert meld_op.EES is None

    meld_op.fit_transform(D, RES)
    assert meld_op.filt is not None
    assert meld_op.EES is not None

    # check changing graph params resets filter
    meld_op.set_params(knn=meld_op.knn + 1)
    assert meld_op.graph is None
    assert meld_op.filt is None
    assert meld_op.EES is None


def test_meld_invalid_lap_type():
    D = np.random.normal(0, 2, (1000, 2))
    # lap type TypeError
    lap_type = "hello world"
    with assert_raises_message(
        ValueError,
        "lap_type value {} not recognized. "
        "Choose from ['combinatorial', 'normalized']".format(lap_type),
    ):
        meld.MELD(verbose=0, lap_type=lap_type).fit(D)


def test_meld_res_wrong_shape():
    D = np.random.normal(0, 2, (100, 2))
    # RES wrong shape
    RES = np.ones([101, 2])
    with assert_raises_message(
        ValueError,
        "Input data ({}) and input graph ({}) "
        "are not of the same size".format(RES.shape, D.shape[0]),
    ):
        meld.MELD(verbose=0).fit_transform(
            X=D, RES=RES,
        )

def test_meld_res_dataframe():
    D = np.random.normal(0, 2, (100, 2))
    # RES wrong shape
    index = pd.Index(['cell_{}'.format(i) for i in range(100)])
    columns = pd.Index(['A'])
    RES = pd.DataFrame(np.ones([100, 1]),
            index=index,
            columns=columns)
    EES = meld.MELD(verbose=0).fit_transform(
        X=D, RES=RES,
    )
    assert np.all(EES.index == index)
    assert np.all(EES.columns == columns)

def test_meld_res_non_numeric():
    D = np.random.normal(size=(100,2))
    RES = np.random.choice(['A'], size=100)
    meld.MELD().fit_transform(D, RES)

    RES = np.random.choice(['A', 'B'], size=100)
    meld.MELD().fit_transform(D, RES)

    RES = np.random.choice(['A', 'B', 'C'], size=100)
    meld.MELD().fit_transform(D, RES)

class TestCluster(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # VertexFrequencyCluster
        # Custom window sizes
        self.window_sizes = np.array([2, 4, 8, 24])
        self.data, self.labels = make_batches(n_pts_per_cluster=100)
        meld_op = meld.MELD(verbose=0)
        self.EES = meld_op.fit_transform(
            self.data, sample_idx=self.labels, RES=self.labels
        )
        self.G = meld_op.graph

    def test_cluster(self):
        vfc_op = meld.VertexFrequencyCluster(window_sizes=self.window_sizes)
        spectrogram = vfc_op.fit_transform(self.G, RES=self.labels, EES=self.EES)
        # test sparse window
        for t in self.window_sizes:
            np.testing.assert_allclose(
                vfc_op._compute_window(self.G.diff_op, t).toarray(),
                vfc_op._compute_window(self.G.diff_op.toarray(), t),
            )
        # test sparse spectrogram
        for window in vfc_op.windows:
            np.testing.assert_allclose(
                vfc_op._compute_spectrogram(self.labels, window),
                vfc_op._compute_spectrogram(self.labels, sparse.csr_matrix(window)),
            )
        # test full sparse computation
        vfc_op.sparse = True
        sparse_spectrogram = vfc_op.fit_transform(self.G, RES=self.labels, EES=self.EES)
        assert sparse_spectrogram.shape == spectrogram.shape
        assert sparse.issparse(vfc_op._basewindow)

    def test_cluster_no_EES(self):
        vfc_op = meld.VertexFrequencyCluster(window_sizes=self.window_sizes)
        spectrogram = vfc_op.fit_predict(self.G, RES=self.labels, EES=None)

    def test_predit_setting_n_cluster(self):
        vfc_op = meld.VertexFrequencyCluster(window_sizes=self.window_sizes)
        spectrogram = vfc_op.fit_transform(self.G, RES=self.labels, EES=None)
        clusters = vfc_op.predict(n_clusters=2)

    def test_2d(self):
        RES = np.array([self.labels, self.labels]).T
        vfc_op = meld.VertexFrequencyCluster(window_sizes=self.window_sizes)
        meld_op = meld.MELD(verbose=0,)
        EES = meld_op.fit_transform(self.data, RES=RES)
        clusters = vfc_op.fit_predict(self.G, RES=RES, EES=EES)
        assert len(clusters) == len(self.labels)

    def test_RES_EES_shape(self):
        RES = np.array([self.labels, self.labels]).T
        vfc_op = meld.VertexFrequencyCluster(window_sizes=self.window_sizes)
        meld_op = meld.MELD(verbose=0,)
        EES = meld_op.fit_transform(self.data, RES=RES)
        assert_raises_message(
            ValueError,
            "`RES` and `EES` must have the same shape."
            "Got RES: {} and EES: {}".format(str(RES[:, 1].shape), str(EES.shape)),
            vfc_op.fit_predict,
            G=self.G,
            RES=RES[:, 1],
            EES=EES,
        )

    def test_transform_before_fit(self):
        # Transform before fit
        assert_raises_message(
            ValueError,
            "Estimator must be `fit` before running `transform`.",
            meld.VertexFrequencyCluster().transform,
            RES=self.labels,
            EES=self.EES,
        )

    def test_predict_before_fit(self):
        # predict before fit
        assert_raises_message(
            ValueError,
            "Estimator is not fit. Call VertexFrequencyCluster.fit().",
            meld.VertexFrequencyCluster().predict
        )

    def test_predict_before_transform(self):
        vfc_op = meld.VertexFrequencyCluster(window_sizes=self.window_sizes)
        vfc_op.fit(self.G)
        # predict before transform
        assert_raises_message(
            ValueError,
            "Estimator is not transformed. " "Call VertexFrequencyCluster.transform().",
            vfc_op.predict
        )

    def test_res_invalid(self):
        # RES not array-like
        assert_raises_message(
            TypeError,
            "`RES` must be array-like",
            meld.VertexFrequencyCluster().fit_transform,
            G=self.G,
            RES="invalid",
            EES=self.EES,
        )

    def test_ees_invalid(self):
        # EES not array-like
        assert_raises_message(
            TypeError,
            "`EES` must be array-like",
            meld.VertexFrequencyCluster().fit_transform,
            G=self.G,
            RES=self.labels,
            EES="invalid",
        )

    def test_res_wrong_length(self):
        # RES and n mismatch
        assert_raises_message(
            ValueError,
            "At least one axis of `RES` must be of length `N`.",
            meld.VertexFrequencyCluster().fit_transform,
            G=self.G,
            RES=np.ones(7),
            EES=self.EES,
        )

    def test_ees_wrong_length(self):
        # EES and n mismatch
        assert_raises_message(
            ValueError,
            "At least one axis of `EES` must be of length `N`.",
            meld.VertexFrequencyCluster().fit_transform,
            G=self.G,
            RES=self.labels,
            EES=np.ones(7),
        )

    def test_set_params(self):
        # KMeans params
        vfc_op = meld.VertexFrequencyCluster()
        vfc_op.set_kmeans_params(n_clusters=2)
        assert vfc_op.n_clusters == 2

    def test_power_sparse(self):
        vfc_op = meld.VertexFrequencyCluster()
        vfc_op._power_matrix(self.G.diff_op, 2)



# Copyright (C) 2020 Krishnaswamy Lab, Yale University

import numpy as np
import graphtools.base
import graphtools
import pygsp
import scprep
import sklearn

def _check_pygsp_graph(G):
    if isinstance(G, graphtools.base.BaseGraph):
        if not isinstance(G, pygsp.graphs.Graph):
            G = G.to_pygsp()
    else:
        raise TypeError(
            "Input graph should be of type graphtools.base.BaseGraph."
            " With graphtools, use the `use_pygsp=True` flag."
        )
    return G


def get_meld_cmap():
    """Returns cmap used in publication for displaying EES.
       Inspired by cmocean `balance` cmap"""
    base_colors = [
        [0.22107637, 0.53245276, 0.72819301, 1.0],
        [0.7, 0.7, 0.7, 1],
        [0.75013244, 0.3420382, 0.22753009, 1.0],
    ]

    return scprep.plot.tools.create_colormap(base_colors)


def normalize_EES_within_replicate(EES):
    '''
    Takes a two-column EES representing matched experimental and control
    pairs from the same replicate and normalizes the EES row-sum to 1.
    '''
    if EES.shape[1] != 2:
        raise ValueError(
            "It is currently only possible to normalize two samaples"
            "from a matched experimental and control pair. As such,"
            "EES.shape[1] must equal 2. Got {}".format(EES.shape[1])
        )
    return sklearn.preprocessing.normalize(EES, norm='l1')



# Copyright (C) 2020 Krishnaswamy Lab, Yale University

import os
import sys
from setuptools import setup, find_packages

install_requires = [
    "numpy>=1.14.0",
    "scipy>=1.1.0",
    "graphtools>=1.5.0",
    "pandas>=0.25",
    "scprep>=1.0",
    "pygsp",
    "sklearn"
]

test_requires = [
    "nose",
    "nose2",
    "coverage",
    "coveralls",
    "scikit-learn",
    "packaging",
    "matplotlib",
    "parameterized"
    ]

doc_requires = [
    "sphinx<=1.8.5",
    "sphinxcontrib-napoleon",
    "autodocsumm",
]

version_py = os.path.join(os.path.dirname(__file__), "meld", "version.py")
version = open(version_py).read().strip().split("=")[-1].replace('"', "").strip()

readme = open("README.md").read()

setup(
    name="meld",
    version=version,
    description="MELD",
    author="Daniel Burkhardt, Krishnaswamy Lab, Yale University",
    author_email="daniel.burkhardt@yale.edu",
    packages=find_packages(),
    license="Dual License - See LICENSE file",
    python_requires=">=3.5",
    install_requires=install_requires,
    extras_require={"test": test_requires, "doc": doc_requires},
    test_suite="nose2.collector.collector",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/KrishnaswamyLab/MELD",
    download_url="https://github.com/KrishnaswamyLab/MELD/archive/v{}.tar.gz".format(
        version
    ),
    keywords=["big-data", "manifold-learning", "computational-biology"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Framework :: Jupyter",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)

# get location of setup.py
setup_dir = os.path.dirname(os.path.realpath(__file__))



# Copyright (C) 2020 Krishnaswamy Lab, Yale University

import numpy as np
import pandas as pd
from scipy import sparse
from sklearn.base import BaseEstimator
from sklearn.cluster import KMeans
from sklearn import preprocessing, decomposition
import warnings

import scprep

import time

from . import utils


class VertexFrequencyCluster(BaseEstimator):
    """Performs Vertex Frequency clustering for data given a
       raw experimental signal and enhanced experimental signal.

    Parameters
    ----------
    n_clusters : int, optional, default: 10
        The number of clusters to form.
    EES_bias : float, optional, default: 1
        A normalization term that biases clustering towards the
        EES (higher values) or towards the spectrogram (lower values)
    window_count : int, optional, default: 9
        Number of windows to use if window_sizes = None
    window_sizes : None, optional, default: None
        ndarray of integer window sizes to supply to t
    sparse : bool, optional, default: False
        Use sparse matrices. This is significantly slower,
        but will use less memory
    suppress : bool, optional
        Suppress warnings
    random_state : int or None, optional (default: None)
        Random seed for clustering
    **kwargs
        Description

    Raises
    ------
    NotImplementedError
        Window functions are not implemented

    Examples
    --------

    """

    def __init__(
        self,
        n_clusters=10,
        EES_bias=1,
        window_count=9,
        window_sizes=None,
        sparse=False,
        suppress=False,
        random_state=None,
        **kwargs
    ):
        self.suppress = suppress
        self.sparse = sparse
        self._basewindow = None
        if window_sizes is None:
            self.window_sizes = np.power(2, np.arange(window_count))
        else:
            self.window_sizes = window_sizes

        self.window_count = np.min(self.window_sizes.shape)
        self.n_clusters = n_clusters
        self.EES_bias = EES_bias
        self.window = None
        self.eigenvectors = None
        self.N = None
        self.spec_hist = None
        self.spectrogram = None
        self.combined_spectrogram_ees = None
        self.isfit = False
        self.EES = None
        self.RES = None
        self._sklearn_params = kwargs

    def _activate(self, x, alpha=1):
        """Activate spectrograms for clustering

        Parameters
        ----------
        x : numeric
            input signal
        alpha : int, optional
            amount of activation

        Returns
        -------
        activated signal
        """
        return np.tanh(alpha * np.abs(x))

    def _compute_spectrogram(self, RES, window):
        """Computes spectrograms for arbitrary window/signal/graph combinations

        Parameters
        ----------
        RES : np.ndarray
            Input signal
        U : np.ndarray
            eigenvectors
        window : TYPE
            window matrix

        Returns
        -------
        C
            Normalized Spectrogram

        Raises
        ------
        TypeError
            Description
        """
        tic = time.time()
        # print('    Computing spectrogram for window')
        if len(RES.shape) == 1:
            RES = RES[:, None]
        if sparse.issparse(window):
            # the next computation becomes dense - better to make dense now
            C = window.multiply(RES).toarray()
        else:
            C = np.multiply(window, RES)
        C = preprocessing.normalize(self.eigenvectors.T @ C, axis=0)
        # print('     finished in {:.2f} seconds'.format(time.time() - tic))
        return C.T

    def _compute_multiresolution_spectrogram(self, RES):
        """ Compute multiresolution spectrogram by repeatedly calling
            _compute_spectrogram """

        # spectrogram = np.zeros((self.windows[0].shape[1],
        #                        self.eigenvectors.shape[1]))

        # for window in self.windows:
        #    curr_spectrogram = self._compute_spectrogram(
        #        RES, window)
        #    curr_spectrogram = self._activate(curr_spectrogram)
        #    spectrogram += curr_spectrogram
        tic = time.time()
        # print('  Computing multiresolution spectrogram')
        spectrogram = np.zeros((self.windows[0].shape[1], self.eigenvectors.shape[1]))

        for window in self.windows:
            curr_spectrogram = self._compute_spectrogram(RES=RES, window=window)
            curr_spectrogram = self._activate(curr_spectrogram)
            spectrogram += curr_spectrogram

        # print('   finished in {:.2f} seconds'.format(time.time() - tic))

        return spectrogram

    def _compute_window(self, window, t=1):
        """_compute_window
        These windows mask the signal (RES) to perform a Windowed Graph
        Fourier Transform (WGFT) as described by Shuman et al.
        (https://arxiv.org/abs/1307.5708).

        This function is used when the power of windows is NOT diadic
        """
        if sparse.issparse(window):
            window = window ** t
        else:
            window = np.linalg.matrix_power(window, t)
        return preprocessing.normalize(window, "l2", axis=0).T

    def _power_matrix(self, a, n):
        if sparse.issparse(a):
            a = a ** n
        else:
            a = np.linalg.matrix_power(a, n)
        return a

    def _compute_windows(self):
        """_compute_window
        These windows mask the signal (RES) to perform a Windowed Graph
        Fourier Transform (WGFT) as described by Shuman et al.
        (https://arxiv.org/abs/1307.5708).

        This function is used when the power of windows is diadic and
        computes all windows efficiently.
        """
        windows = []
        curr_window = self._basewindow
        windows.append(preprocessing.normalize(curr_window, "l2", axis=0).T)
        for i in range(len(self.window_sizes) - 1):
            curr_window = self._power_matrix(curr_window, 2)
            windows.append(preprocessing.normalize(curr_window, "l2", axis=0).T)
        return windows

    def _combine_spectrogram_EES(self, spectrogram, EES):
        """ Normalizes and concatenates the EES to the
            spectrogram for clustering"""

        spectrogram_n = spectrogram / np.linalg.norm(spectrogram)

        ees_n = EES / np.linalg.norm(EES, ord=2, axis=0)
        ees_n = ees_n * self.EES_bias
        data_nu = np.c_[spectrogram_n, ees_n]
        return data_nu

    def fit(self, G):
        """Sets eigenvectors and windows."""

        self.graph = utils._check_pygsp_graph(G)

        if self._basewindow is None:
            self._basewindow = G.diff_op
        if not self.sparse and sparse.issparse(self._basewindow):
            self._basewindow = self._basewindow.toarray()
        elif self.sparse and not sparse.issparse(self._basewindow):
            self._basewindow = sparse.csr_matrix(self._basewindow)

        self.windows = []
        tic = time.time()
        # print('Building windows')
        # Check if windows were generated using powers of 2
        if np.all(np.diff(np.log2(self.window_sizes)) == 1):
            self.windows = self._compute_windows()
        else:
            for t in self.window_sizes:
                self.windows.append(
                    self._compute_window(self._basewindow, t=t).astype(float)
                )
        # print(' finished in {:.2f} seconds'.format(time.time() - tic))

        tic = time.time()
        # print('Computing Fourier basis')
        # Compute Fourier basis. This may take some time.
        self.graph.compute_fourier_basis()
        self.eigenvectors = self.graph.U
        self.N = self.graph.N
        self.isfit = True
        # print(' finished in {:.2f} seconds'.format(time.time() - tic))

        return self

    def transform(self, RES, EES=None, center=True):
        """Calculates the spectrogram of the graph using the RES"""
        self.RES = RES
        self.EES = EES
        if not self.isfit:
            raise ValueError("Estimator must be `fit` before running `transform`.")

        if not isinstance(self.RES, (list, tuple, np.ndarray, pd.Series, pd.DataFrame)):
            raise TypeError("`RES` must be array-like.")

        if EES is not None and not isinstance(
            self.EES, (list, tuple, np.ndarray, pd.Series)
        ):
            raise TypeError("`EES` must be array-like.")

        # Checking shape of RES
        self.RES = np.array(self.RES)
        if not self.N in self.RES.shape:
            raise ValueError("At least one axis of `RES` must be" " of length `N`.")

        # Checking shape of EES
        if EES is not None:
            if self.N not in self.EES.shape:
                raise ValueError("At least one axis of `EES` must be" " of length `N`.")
            if EES.shape != RES.shape:
                raise ValueError(
                    "`RES` and `EES` must have the same shape."
                    "Got RES: {} and EES: {}".format(str(RES.shape), str(EES.shape))
                )
            self.EES = np.array(self.EES)

        # Subtract the mean from the RES
        if center:
            self.RES = self.RES - self.RES.mean()

        # If only one RES, no need to collect
        if len(self.RES.shape) == 1:
            self.spectrogram = self._compute_multiresolution_spectrogram(self.RES)
        else:
            # Create a list of spectrograms and concatenate them
            spectrograms = []
            for i in range(self.RES.shape[1]):
                curr_RES = scprep.select.select_cols(self.RES, idx=i)
                spectrograms.append(self._compute_multiresolution_spectrogram(curr_RES))
            self.spectrogram = np.hstack(spectrograms)

        # Appending the EES to the spectrogram
        if self.EES is not None:
            self.combined_spectrogram_ees = self._combine_spectrogram_EES(
                spectrogram=self.spectrogram, EES=self.EES
            )

        return self.spectrogram

    def fit_transform(self, G, RES, EES=None, **kwargs):
        self.fit(G, **kwargs)
        return self.transform(RES, EES, **kwargs)

    def predict(self, n_clusters=None, **kwargs):
        """Runs KMeans on the spectrogram."""

        if n_clusters is not None:
            self.n_clusters = n_clusters
        self._clusterobj = KMeans(n_clusters=self.n_clusters, **kwargs, **self._sklearn_params)

        if not self.isfit:
            raise ValueError(
                "Estimator is not fit. " "Call VertexFrequencyCluster.fit()."
            )
        if self.spectrogram is None:
            raise ValueError(
                "Estimator is not transformed. "
                "Call VertexFrequencyCluster.transform()."
            )

        if self.combined_spectrogram_ees is None:
            data = self.spectrogram
        else:
            data = self.combined_spectrogram_ees
        tic = time.time()
        # print('Running PCA on the spectrogram')
        data = decomposition.PCA(self.n_clusters).fit_transform(data)
        # print(' finished in {:.2f} seconds'.format(time.time()-tic))

        tic = time.time()
        # print('Running clustering')
        self.labels_ = self._clusterobj.fit_predict(data)

        self.labels_ = scprep.utils.sort_clusters_by_values(self.labels_, self.RES)
        # print(' finished in {:.2f} seconds'.format(time.time()-tic))

        return self.labels_

    def fit_predict(self, G, RES, EES=None, **kwargs):
        self.fit_transform(G, RES, EES, **kwargs)
        return self.predict()

    def set_kmeans_params(self, **kwargs):
        k = kwargs.pop("n_clusters", False)
        if k:
            self.n_clusters = k
        self._sklearn_params = kwargs
